     /**
      * \brief JSON Service program
      *
      * From the main web site:
      *
      * <p>
      * JSON (JavaScript Object Notation) is a lightweight data-interchange
      * format. It is easy for humans to read and write. It is easy for
      * machines to parse and generate. It is based on a subset of the
      * JavaScript Programming Language, Standard ECMA-262 3rd Edition -
      * December 1999. JSON is a text format that is completely language
      * independent but uses conventions that are familiar to programmers
      * of the C-family of languages, including C, C++, C#, Java, JavaScript,
      * Perl, Python, and many others. These properties make JSON an ideal
      * data-interchange language.
      * </p>
      *
      * JSON is built on two structures:
      *
      * <ul>
      *   <li>A collection of name/value pairs. In various languages, this is
      *       realized as an object, record, struct, dictionary, hash table,
      *       keyed list, or associative array.
      *   </li>
      *   <li>
      *       An ordered list of values. In most languages, this is realized as
      *       an array, vector, list, or sequence.
      *   </li>
      * </ul>
      *
      * <p>
      * These are universal data structures. Virtually all modern programming
      * languages support them in one form or another. It makes sense that a
      * data format that is interchangable with programming languages also be
      * based on these structures.
      * </p>
      *
      * In JSON, they take on these forms:
      *
      * <p>
      * An object is an unordered set of name/value pairs. An object begins
      * with { (left brace) and ends with } (right brace). Each name is followed
      * by : (colon) and the name/value pairs are separated by , (comma).
      * </p>
      *
      * Concerning this implementation:
      *
      * <br><br>
      *
      * The pointer to the JSON object will be check on every procedure call.
      * If the pointer does not point to a JSON object an escape message will
      * be send to the next level of the call stack.
      *
      * <br><br>
      *
      * If the JSON obect is queried for an entry of a specific type and the
      * type does not match with the expectation of the procedure an escape
      * message will be send to the next level of the call stack.
      *
      * <br><br>
      *
      * All the allocated memory of the JSON object and its entries is managed
      * by this service program and will be freed with the call of the
      * procedure <em>dispose</em>. No manual deallocation must be done.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \link http://www.json.org JSON
      *
      * \rev 27.01.2010 Mihael Schmidt
      *      Added support for getting numbers with procedures json_getInt,
      *      json_getLong, json_getDouble no matter what the original type was
      *      (f. e. double value will be cast to int with json_getInt).
      *
      * \rev 13.02.2011 Mihael Schmidt
      *      Fixed bug in remove procedure. Replaced string and hex constants
      *      with unicode constants.
      * 
      * \rev 09.04.2011 Mihael Schmidt
      *      Fixed memory management error in toString procedure. 
      * 
      * \rev 24.06.2011 Brian Garland
      *      Remove leading plus (+) from floating point numbers when creating
      *      a string from a json object.
      * 
      * \rev 23.10.2011 Mihael Schmidt
      *      Switched from Linked List to Arraylist and by that removing a 
      *      dependency to Linked List <br>
      *      Fixed bug in escapeLastToken procedure
      * 
      */

      *-------------------------------------------------------------------------
      * Copyright (c) 2007-2011 Mihael Schmidt
      * All rights reserved.
      *
      * This file is part of the JSON service program.
      *
      * JSON service program is free software: you can redistribute it and/or modify it under
      * the terms of the GNU Lesser General Public License as published by
      * the Free Software Foundation, either version 3 of the License, or
      * any later version.
      *
      * JSON service program is distributed in the hope that it will be useful,
      * but WITHOUT ANY WARRANTY; without even the implied warranty of
      * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
      * GNU Lesser General Public License for more details.
      *
      * You should have received a copy of the GNU Lesser General Public
      * License along with JSON service program.  If not, see http://www.gnu.org/licenses/.
      *-------------------------------------------------------------------------


     H nomain
     H bnddir('QC2LE')
     H copyright('Copyright (c) 2007-2010 Mihael Schmidt. All rights reserved.')

      *-------------------------------------------------------------------------
      * Prototypes
      *-------------------------------------------------------------------------
     D json_srvpgm_init...
     D                 PR            10I 0 extproc('json_srvpgm_init')
      *
      /copy 'json_h.rpgle'
      /copy 'json_int_h.rpgle'
      /copy 'ceeapi_h.rpgle'
      /copy 'libc_h.rpgle'
      /copy 'arraylist_h.rpgle'

      /if not defined (QUSEC)
      /define QUSEC
      /copy QSYSINC/QRPGLESRC,QUSEC
      /endif


      *-------------------------------------------------------------------------
      * Constants
      *-------------------------------------------------------------------------
     D HEX_NULL        C                   x'00'
     D JSON_ID         C                   'JSON_IMPLEMENTATION'

      /copy 'json_c.rpgle'
      /copy 'unicode_c.rpgle'


      *-------------------------------------------------------------------------
      * Global Variables
      *-------------------------------------------------------------------------
      *
      * Parameters passed to CEEDOD
      *
     D descType        S             10I 0
     D dataType        S             10I 0
     D descInfo1       S             10I 0
     D descInfo2       S             10I 0
     D length          S             10I 0
      *
     D chars           DS                  likeds(tmpl_chars)
     D chars_compact   DS                  likeds(tmpl_chars)
      *
     D hexNull         S              1A   inz(HEX_NULL)


      *-------------------------------------------------------------------------
      * Procedures
      *-------------------------------------------------------------------------

     /**
      * \brief Service program initialization
      *
      * String formattings are preloaded.
      */
     P json_srvpgm_init...
     P                 B                   export
     D                 PI            10I 0
      /free
       chars.d_quote = %char(UNICODE_DOUBLE_QUOTE) + x'00';
       chars.l_brace = ' ' + %char(UNICODE_LEFT_CURLY_BRACE) + ' ' + x'00';
       chars.r_brace = ' ' + %char(UNICODE_RIGHT_CURLY_BRACE) + ' ' + x'00';
       chars.l_bracket = ' ' + %char(UNICODE_LEFT_BRACKET) + ' ' + x'00';
       chars.r_bracket = ' ' + %char(UNICODE_RIGHT_BRACKET) + ' ' + x'00';
       chars.comma = ' ' + %char(UNICODE_COMMA) + ' ' + x'00';
       chars.colon = ' ' + %char(UNICODE_COLON) + ' ' + x'00';
       chars.true = 'true' + x'00';
       chars.false = 'false' + x'00';
       chars.null_str = 'null' + x'00';
       chars.size_d_quote = 1;
       chars.size_l_brace = 3;
       chars.size_r_brace = 3;
       chars.size_l_bracket = 3;
       chars.size_r_bracket = 3;
       chars.size_comma = 3;
       chars.size_colon = 3;
       chars.size_true = 4;
       chars.size_false = 5;
       chars.size_null_str = 4;

       chars_compact.d_quote = %char(UNICODE_DOUBLE_QUOTE) + x'00';
       chars_compact.l_brace = %char(UNICODE_LEFT_CURLY_BRACE) + x'00';
       chars_compact.r_brace = %char(UNICODE_RIGHT_CURLY_BRACE) + x'00';
       chars_compact.l_bracket = %char(UNICODE_LEFT_BRACKET) + x'00';
       chars_compact.r_bracket = %char(UNICODE_RIGHT_BRACKET) + x'00';
       chars_compact.comma = %char(UNICODE_COMMA) + x'00';
       chars_compact.colon = %char(UNICODE_COLON) + x'00';
       chars_compact.true = 'true' + x'00';
       chars_compact.false = 'false' + x'00';
       chars_compact.null_str = 'null' + x'00';
       chars_compact.size_d_quote = 1;
       chars_compact.size_l_brace = 1;
       chars_compact.size_r_brace = 1;
       chars_compact.size_l_bracket = 1;
       chars_compact.size_r_bracket = 1;
       chars_compact.size_comma = 1;
       chars_compact.size_colon = 1;
       chars_compact.size_true = 4;
       chars_compact.size_false = 5;
       chars_compact.size_null_str = 4;

       return 0;
      /end-free
     P                 E


     /**
      * \brief Create JSON Object
      *
      * Creates a JSON Object.
      *
      * <br><br>
      *
      * The memory allocated for this object must be deallocated with the
      * <em>dispose</em> procedure when finished.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \return Pointer to the JSON object
      */
     P json_create     B                   export
     D                 PI              *
      *
     D jsonPtr         S               *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
      /free
       jsonPtr = %alloc(%size(tmpl_header));
       header.id = JSON_ID;
       header.size = 0;
       header.firstEntry = *null;
       header.lastEntry = *null;
       header.iteration = -1;
       header.iterNextEntry = *null;
       header.iterPrevEntry = *null;
       header.disposeList = arraylist_create();

       return jsonPtr;
      /end-free
     P                 E


     /**
      * \brief Deallocate JSON object
      *
      * The memory for the JSON Object will be deallocated and the passed
      * pointer will be set to *null.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P json_dispose    B                   export
     D                 PI
     D   jsonPtr                       *
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D ptr             S               *
     D ptrDisposeElement...
     D                 S               *   based(ptr)
     D size            S             10I 0
     D i               S             10I 0
      /free
       if (jsonPtr <> *null);
         isJSONImpl(jsonPtr);

         // deallocate all temporary allocated memory from the dispose list
         size = arraylist_getSize(header.disposeList);
         size -= 1;
         for i = 0 to size;
           // a user may have already freed the memory so ignore all errors
           ptr = arraylist_get(header.disposeList : i);
           monitor;
             if (ptrDisposeElement <> *null);
               dealloc ptrDisposeElement;
             endif;
             on-error *all;
           endmon;
             
         endfor;

         arraylist_dispose(header.disposeList);

         json_clear(jsonPtr);
         dealloc(n) jsonPtr;
       endif;
      /end-free
     P                 E


     /**
      * \brief Remove entry
      *
      * If the JSON object contains such an entry it will be removed.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Entry key
      *
      */
     P json_remove     B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D ptr             S               *
     D entry           DS                  likeds(tmpl_entry) based(ptr)
     D nextEntryPtr    S               *
     D nextEntry       DS                  likeds(tmpl_entry)
     D                                     based(nextEntryPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)

      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       ptr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);

       // check if entry was found
       if (ptr <> *null);

         // update header
         if (header.firstEntry = ptr);
           header.firstEntry = entry.next;
         endif;
         if (header.lastEntry = ptr);
           header.lastEntry = entry.prev;
         endif;

         // update previous and next (if exist)
         if (entry.prev <> *null);
           prevEntryPtr = entry.prev;
           prevEntry.next = entry.next;
         endif;
         if (entry.next <> *null);
           nextEntryPtr = entry.next;
           nextEntry.prev = entry.prev;
         endif;

         // release memory of entry value
         disposeEntryValue(ptr);

         dealloc(n) entry.key;
         dealloc(n) ptr;
         
         header.size -= 1;
       endif;

      /end-free
     P                 E


     /**
      * \brief Clear JSON object
      *
      * Removes all entries from the JSON object.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      */
     P json_clear      B                   export
     D                 PI
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D tmpPtr          S               *
     D ptr             S               *
     D entry           DS                  likeds(tmpl_entry) based(ptr)
      /free
       isJSONImpl(jsonPtr);

       ptr = header.lastEntry;
       dow (ptr <> *null);
         tmpPtr = entry.prev;

         // dispose/dealloc old value
         disposeEntryValue(ptr);

         dealloc(n) entry.key;
         dealloc(n) ptr;
         ptr = tmpPtr;
       enddo;

       // update header
       header.size = 0;
       header.firstEntry = *null;
       header.lastEntry = *null;
       header.iteration = -1;
       header.iterNextEntry = *null;
       header.iterPrevEntry = *null;
      /end-free
     P                 E


     /**
      * \brief Get next entry
      *
      * Iterates through all JSON entries and returns a pointer to the key of
      * the next entry.
      *
      * <br><br>
      *
      * If the iteration ends with the last key *null will be returned.
      *
      * <br><br>
      *
      * If the iteration is aborted the calling program should use the
      * <em>abortIteration</em> procedure to reset the iteration.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      * \return Pointer to the key of the next entry or *null if there are no
      *         more entries
      */
     P json_getNext    B                   export
     D                 PI              *
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *
      /free
       isJSONImpl(jsonPtr);

       if (header.iteration = header.size-1  or
           header.iteration < -1);
         header.iteration = -1;
         retVal = *null;
       else;
         if (header.iterNextEntry = *null);
           entryPtr = header.firstEntry;
         else;
           entryPtr = header.iterNextEntry;
         endif;

         header.iteration += 1;
         header.iterNextEntry = entry.next;
         retVal = entry.key;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Abort iteration
      *
      * Resets the iteration of the JSON Object.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P json_abortIteration...
     P                 B                   export
     D                 PI
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header)  based(jsonPtr)
      /free
       isJSONImpl(jsonPtr);

       header.iteration = -1;
       header.iterNextEntry = *null;
       header.iterPrevEntry = *null;
      /end-free
     P                 E


     /**
      * \brief Contains key
      *
      * Checks if the JSON object contains an entry with the passed key.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Entry key
      *
      * \return *on = JSON object contains key<br>
      *         *off = JSON object does not contain key
      */
     P json_contains...
     P                 B                   export
     D                 PI              N   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D entryPtr        S               *
     D retVal          S               N
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr = *null);
         retVal = *off;
       else;
         retVal = *on;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Is JSON object empty
      *
      * Checks if the JSON object contains no entries at all.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      * \return *on = JSON object is empty <br>
      *         *off = JSON object is not empty
      */
     P json_isEmpty    B                   export
     D                 PI              N
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D retVal          S               N
      /free
       isJSONImpl(jsonPtr);

       if (header.size = 0);
         retVal = *on;
       else;
         retVal = *off;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Count JSON object entries
      *
      * Returns the number of entries in the JSON object. The keys of the
      * entries are counted. If the JSON object contains arrays the elements
      * of the arrays are ignored.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      *
      * \return number of entries in the JSON object
      */
     P json_size       B                   export
     D                 PI            10I 0
     D  jsonPtr                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
      /free
       isJSONImpl(jsonPtr);

       return header.size;
      /end-free
     P                 E


     /**
      * \brief Add character entry
      *
      * Adds an entry of type character.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putString  B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                    65535A   const options(*varsize)
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D tmpValue        S          65535A
     D keyLength       S             10I 0
     D valueLength     S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);
       cee_getOpDescInfo(3 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       valueLength = length;
       tmpValue = %subst(value : 1 : valueLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = valueLength;
         newEntry.value = %alloc(valueLength + 1);   // +1 for the null byte
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_STRING;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = valueLength;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       memcpy(newEntry.value : %addr(tmpValue) : valueLength);

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
       memcpy(newEntry.value + valueLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add boolean entry
      *
      * Adds an entry of type boolean.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putBoolean...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         N   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S               N   based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);


       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_BOOLEAN;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 1;
         newEntry.value = %alloc(1);
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_BOOLEAN;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 1;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;


       // copy key/value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add integer entry
      *
      * Adds an entry of type integer.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putInt...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                       10I 0 const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S             10I 0 based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_INT;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue));
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;
       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_INT;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add long entry
      *
      * Adds an entry of type long.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putLong...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                       20I 0 const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S             20I 0 based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_LONG;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue));
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_LONG;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add double entry
      *
      * Adds an entry of type double.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Value
      */
     P json_putDouble...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                        8F   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
     D tmpValue        S              8F   based(newEntry.value)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_DOUBLE;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = %size(tmpValue);
         newEntry.value = %alloc(%size(tmpValue));
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_DOUBLE;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = %size(tmpValue);

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       tmpValue = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add array entry
      *
      * Adds an entry of type array.
      *
      * \author Mihael Schmidt
      * \date   19.04.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      * \param Array
      */
     P json_putArray...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   value                         *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpKey          S          65535A
     D keyLength       S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_ARRAY;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 0;
         newEntry.value = *null;
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_ARRAY;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 0;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy key and value to entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       newEntry.value = value;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get entry
      *
      * Gets the value of entry in the JSON object as a string.
      *
      * <br><br>
      *
      * The return value is a pointer to a null terminated string so that it
      * can be easily processed with the %str BIF.
      *
      * <br><br>
      *
      * The memory for the return value was allocated will be managed by
      * the JSON object and will be disposed with the <em>dispose</em>
      * procedure. The calling program does not need to deallocate the
      * the memory by itself.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Pointer to a null terminated string or *null if JSON object
      *         does not contain the passed key
      */
     P json_get        B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *   inz(*null)
      *
     D tmpPtr          S               *
     D tmpDs           DS                  based(tmpPtr)
     D   tmpBoolean                    N   overlay(tmpDs)
     D   tmpInt                      10I 0 overlay(tmpDs)
     D   tmpLong                     20I 0 overlay(tmpDs)
     D   tmpDouble                    8F   overlay(tmpDs)
     D tmpString       S            310A
     D tmpLength       S             10I 0
     D hex_null        S              1A   inz(x'00')
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         select;
           when (entry.type = JSON_TYPE_STRING);
             retVal = %alloc(entry.valueLength + 1); // + 1 byte for the null value
             // +1 because the null byte was already added to the string during putString
             memcpy(retVal : entry.value : entry.valueLength + 1);

           when (entry.type = JSON_TYPE_INT);
             tmpPtr = entry.value;
             tmpString = %char(tmpInt);
             tmpLength = %len(%trimr(tmpString));
             retVal = %alloc(tmpLength);
             memcpy(retVal : %addr(tmpString) : tmpLength);
             memcpy(retVal + tmplength : %addr(hex_null) : 1);

           when (entry.type = JSON_TYPE_LONG);
             tmpPtr = entry.value;
             tmpString = %char(tmpLong);
             tmpLength = %len(%trimr(tmpString));
             retVal = %alloc(tmpLength);
             memcpy(retVal : %addr(tmpString) : tmpLength);
             memcpy(retVal + tmplength : %addr(hex_null) : 1);

           when (entry.type = JSON_TYPE_DOUBLE);
             tmpPtr = entry.value;
             tmpString = %char(tmpDouble);
             tmpLength = %len(%trimr(tmpString));
             retVal = %alloc(tmpLength);
             memcpy(retVal : %addr(tmpString) : tmpLength);
             memcpy(retVal + tmplength : %addr(hex_null) : 1);

           when (entry.type = JSON_TYPE_BOOLEAN);
             tmpPtr = entry.value;
             retVal = %alloc(%size(tmpBoolean));
             memcpy(retVal : %addr(tmpDs) : %size(tmpBoolean));
             memcpy(retVal + 1 : %addr(hex_null) : 1);

           when (entry.type = JSON_TYPE_ARRAY);
             retVal = jsona_toString(entry.value);

           when (entry.type = JSON_TYPE_OBJECT);
             retVal = json_toString(entry.value);

         endsl;
       endif;

       // register this pointer to the dispose list so that the memory will be managed
       if (retVal <> *null);
         if (entry.type = JSON_TYPE_OBJECT);
           // don't add this pointer to the dispose list
           // it has already been added to the list in the toString procedure
         else;
           arraylist_add(header.disposeList : %addr(retVal) : 16);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get string entry
      *
      * Returns a pointer to the value of the entry for the passed key.
      * The value is null terminated and can be processed with the %str BIF.
      *
      * <br><br>
      *
      * If the entry is not of type string an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Pointer to a null-terminated string
      */
     P json_getString  B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *   inz(*null)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_STRING);
           retVal = entry.value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get boolean entry
      *
      * Returns the boolean value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type boolean an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Boolean value of the entry
      */
     P json_getBoolean...
     P                 B                   export
     D                 PI              N   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D value           S               N   based(entry.value)
     D retVal          S               N
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_BOOLEAN);
           retVal = value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get integer entry
      *
      * Returns the integer value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type integer an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Integer value of the entry
      */
     P json_getInt...
     P                 B                   export
     D                 PI            10I 0 opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D valueInt        S             10I 0 based(entry.value)
     D valueLong       S             20I 0 based(entry.value)
     D valueDouble     S              8F   based(entry.value)
     D retVal          S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_INT);
           retVal = valueInt;
         elseif (entry.type = JSON_TYPE_LONG);
           monitor;
             retVal = valueLong;
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         elseif (entry.type = JSON_TYPE_DOUBLE);
           monitor;
             retVal = %int(valueDouble);
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get long entry
      *
      * Returns the long value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type long an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Long value of the entry
      */
     P json_getLong...
     P                 B                   export
     D                 PI            20I 0 opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D valueInt        S             10I 0 based(entry.value)
     D valueLong       S             20I 0 based(entry.value)
     D valueDouble     S              8F   based(entry.value)
     D retVal          S             20I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_LONG);
           retVal = valueLong;
         elseif (entry.type = JSON_TYPE_INT);
           retVal = valueInt;
         elseif (entry.type = JSON_TYPE_DOUBLE);
           monitor;
             retVal = %int(valueDouble);
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get double entry
      *
      * Returns the double value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type double an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Double value of the entry
      */
     P json_getDouble...
     P                 B                   export
     D                 PI             8F   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D valueInt        S             10I 0 based(entry.value)
     D valueLong       S             20I 0 based(entry.value)
     D valueDouble     S              8F   based(entry.value)
     D retVal          S              8F
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_DOUBLE);
           retVal = valueDouble;
         elseif (entry.type = JSON_TYPE_INT);
           retVal = %float(valueInt);
         elseif (entry.type = JSON_TYPE_LONG);
           monitor;
             retVal = %float(valueLong);
             on-error *all;
               sendEscapeMessage(MSG_TYPE_CAST_EXCEPTION);
           endmon;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get array entry
      *
      * Returns the array value of the entry.
      *
      * <br><br>
      *
      * If the entry is not of type array an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Pointer to the JSON array
      */
     P json_getArray...
     P                 B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_ARRAY);
           retVal = entry.value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief String representation of the JSON object in JSON syntax
      *
      * Creates a null terminated string representation of the JSON object in
      * JSON syntax.
      *
      * <br><br>
      *
      * For entries of type string double quotes and backslashes will be escaped
      * (" => \" and \ => \\).
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P json_toString   B                   export
     D                 PI              *
     D   jsonPtr                       *   const
     D   pOutputFormat...
     D                               10I 0 const options(*nopass)
     D   pPrettyPrintFormat...
     D                                     likeds(json_prettyPrintFormat)
     D                                     const options(*nopass)
     D   pIndentationLevel...
     D                               10I 0 options(*nopass)
      *
     D outputFormat    S             10I 0
     D prettyPrintOption...
     D                 S               N   inz(*off)
     D prettyPrintFormat...
     D                 DS                  likeds(json_prettyPrintFormat) inz
     D indentationLevel...
     D                 S             10I 0 inz(0)
      *
     D l_chars         DS                  likeds(tmpl_chars) inz
      *
     D json_string_ptr...
     D                 S               *
     D json_string_length...
     D                 S             10I 0
     D tokenList       S               *
      *
     D tmpDs           DS                  based(entry.value)
     D   tmpBoolean                    N   overlay(tmpDs)
     D   tmpInt                      10I 0 overlay(tmpDs)
     D   tmpLong                     20I 0 overlay(tmpDs)
     D   tmpDouble                    8F   overlay(tmpDs)
     D   tmpArray                      *   overlay(tmpDs)
     D tmpString       S            310A
     D tmpLength       S             10I 0
     D hex_null        S              1A   inz(x'00')
     D newline         S              2A   inz(x'0D25')
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
      *
     D ptr             S               *
     D x               S             10I 0
     D i               S             10I 0
     D size            S             10I 0
      /free
       isJSONImpl(jsonPtr);

       if (%parms() = 2);
         outputFormat = pOutputFormat;
         prettyPrintFormat.useTabs = *off;
         prettyPrintFormat.numberWhitespace = 4;
         prettyPrintFormat.breakAfterElement = *on;
         prettyPrintFormat.breakAfterObjectBrace = *on;
         prettyPrintFormat.breakBeforeArrayBrace = *on;
         prettyPrintFormat.breakAfterArrayBrace = *on;
       elseif (%parms() = 3);
         outputFormat = pOutputFormat;
         prettyPrintFormat = pPrettyPrintFormat;
       elseif (%parms() = 4);
         outputFormat = pOutputFormat;
         prettyPrintFormat = pPrettyPrintFormat;
         indentationLevel = pIndentationLevel;
       endif;

       if (outputFormat = JSON_OUTPUT_PRETTY_PRINT);
         prettyPrintOption = *on;
       endif;

       exsr init;

       // initialisierung
       //  - anfangszeichen schreiben
       //  - tokenlist initialisieren
       tokenList = arraylist_create();
       arraylist_add(tokenList : %addr(l_chars.l_brace) : 
                     %size(l_chars.size_l_brace));

       if (prettyPrintOption);

         if (prettyPrintFormat.breakAfterObjectBrace);
           indentationLevel += 1;
           arraylist_add(tokenList : %addr(newline) : %size(newline));
         endif;

       endif;

       // eintrge verarbeiten
       entryPtr = header.firstEntry;
       dow (entryPtr <> *null);

         if (prettyPrintOption);
           prettyPrint(tokenList : prettyPrintFormat : indentationLevel);
         endif;

         // add key as token to token list
         arraylist_add(tokenList : %addr(l_chars.d_quote) : 
                       l_chars.size_d_quote);
         arraylist_add(tokenList : entry.key : entry.keyLength -1);
         arraylist_add(tokenList : %addr(l_chars.d_quote) : 
                       l_chars.size_d_quote);

         // add key/value separator
         arraylist_add(tokenList : %addr(l_chars.colon) : l_chars.size_colon);

         select;
           when (entry.type = JSON_TYPE_STRING);
             arraylist_add(tokenList : %addr(l_chars.d_quote) : 
                           l_chars.size_d_quote);
             // +1 weil das null byte schon beim hinzufgen des wertes hinzugefgt wurde
             arraylist_add(tokenList : entry.value : entry.valueLength + 1);
             escapeLastToken(tokenList);
             arraylist_add(tokenList : %addr(l_chars.d_quote) : 
                           l_chars.size_d_quote);

           when (entry.type = JSON_TYPE_INT);
             tmpString = %char(tmpInt);
             tmpLength = %len(%trimr(tmpString));
             arraylist_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = JSON_TYPE_LONG);
             tmpString = %char(tmpLong);
             tmpLength = %len(%trimr(tmpString));
             arraylist_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = JSON_TYPE_DOUBLE);
             tmpString = %triml(%char(tmpDouble):'+');
             tmpLength = %len(%trimr(tmpString));
             // JSON Syntax erwartet immer ein . als Trennzeichen (und kein ,)
             tmpString = %xlate(',' : '.' : tmpString);
             arraylist_add(tokenList : %addr(tmpString) : tmpLength);

           when (entry.type = JSON_TYPE_BOOLEAN);
             if (tmpBoolean);
               arraylist_add(tokenList : %addr(l_chars.true) : 
                             l_chars.size_true);
             else;
               arraylist_add(tokenList : %addr(l_chars.false) : 
                             l_chars.size_false);
             endif;

           when (entry.type = JSON_TYPE_ARRAY);
             ptr = jsona_toString(entry.value : outputFormat:prettyPrintFormat:
                                  indentationLevel);
             arraylist_add(tokenList : ptr : strlen(ptr));
             dealloc(n) ptr;

           when (entry.type = JSON_TYPE_OBJECT);
             ptr = json_toString(entry.value : outputFormat : prettyPrintFormat:
                                 indentationLevel);
             arraylist_add(tokenList : ptr : strlen(ptr));

           when (entry.type = JSON_TYPE_NULL);
             arraylist_add(tokenList : %addr(l_chars.null_str) :
                           l_chars.size_null_str);

         endsl;


         // komma hinzufgen falls es nicht der letzte eintrag ist
         if (entryPtr <> header.lastEntry);
           arraylist_add(tokenList : %addr(l_chars.comma) : l_chars.size_comma);

           if (prettyPrintOption and prettyPrintFormat.breakAfterElement);
             arraylist_add(tokenList : %addr(newline) : %size(newline));
           endif;
         endif;

         // nchsten eintrag
         entryPtr = entry.next;
       enddo;

       if (prettyPrintOption);
         if (prettyPrintFormat.breakAfterObjectBrace);
           arraylist_add(tokenList : %addr(newline) : %size(newline));
           indentationLevel -= 1;
         endif;
         prettyPrint(tokenList : prettyPrintFormat : indentationLevel);
       endif;

       // endezeichen schreiben falls eintrge vorhanden
       arraylist_add(tokenList : %addr(l_chars.r_brace) : l_chars.size_r_brace);

       //  json string lnge ermitteln
       // calculate json string length
       size = %int(arraylist_getSize(tokenList)) - 1;
       for i = 0 to size;
         ptr = arraylist_get(tokenList : i);
         json_string_length += strlen(ptr);
       endfor;
       // +1 fr hex Null
       json_string_length += 1;

       // speicher fr json string reservieren
       json_string_ptr = %alloc(json_string_length);

       // build json string
       size = %int(arraylist_getSize(tokenList)) - 1;
       for i = 0 to size;
         ptr = arraylist_get(tokenList : i);
         tmpLength = strlen(ptr);
         memcpy(json_string_ptr + x : ptr : tmpLength);
         x += tmpLength;
       endfor;

       // hex Null anhngen
       memcpy(json_string_ptr + x : %addr(hexNull) : 1);

       // aufrumen
       if (tokenList <> *null);
         arraylist_dispose(tokenList);
       endif;

       // register allocated memory at the dispose list
       if (json_string_ptr <> *null);
         arraylist_add(header.disposeList : %addr(json_string_ptr) : 
                       %size(json_string_ptr));
       endif;

       return json_string_ptr;

       begsr init;
         if (outputFormat = JSON_OUTPUT_COMPACT);
           l_chars = chars_compact;
         else;
           l_chars = chars;
         endif;
       endsr;
      /end-free
     P                 E


     /**
      * \brief Add null value entry
      *
      * Adds an entry of type null to the array.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Key
      */
     P json_putNull    B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpPtr          S               *
     D tmpKey          S          65535A
     D keyLength       S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_NULL;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 0;
         newEntry.value = *null;
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_NULL;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 0;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;


       // copy key/value to the list entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Add JSON object entry
      *
      * Adds an entry of type object to the array.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Key
      * \param Pointer to the JSON object
      */
     P json_putObject...
     P                 B                   export
     D                 PI                  opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
     D   objPtr                        *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D prevEntryPtr    S               *
     D prevEntry       DS                  likeds(tmpl_entry)
     D                                     based(prevEntryPtr)
     D newEntryPtr     S               *
     D newEntry        DS                  likeds(tmpl_entry)
     D                                     based(newEntryPtr)
     D tmpKey          S          65535A
     D keyLength       S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       keyLength = length;
       tmpKey = %subst(key : 1 : keyLength);

       // check if key is in already in the map
       newEntryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : keyLength);
       if (newEntryPtr = *null);    // entry not in json object
         // create new entry
         newEntryPtr = %alloc(%size(tmpl_entry));
         newEntry.type = JSON_TYPE_OBJECT;
         newEntry.keyLength = keyLength + 1;      // +1 for the null byte
         newEntry.key = %alloc(keyLength + 1);     // +1 for the null byte
         newEntry.valueLength = 0;
         newEntry.value = *null;
         newEntry.next = *null;
         newEntry.prev = *null;

         // update header
         header.size += 1;
         if (header.firstEntry = *null);
           header.firstEntry = newEntryPtr;
         else;
           prevEntryPtr = header.lastEntry;
           prevEntry.next = newEntryPtr;
           newEntry.prev = prevEntryPtr;
         endif;
         header.lastEntry = newEntryPtr;

       else;
         // entry needs to be resized
         newEntry.type = JSON_TYPE_OBJECT;
         newEntry.keyLength = keyLength + 1;
         newEntry.key = %realloc(newEntry.key : keyLength + 1);
         newEntry.valueLength = 0;

         // dispose/dealloc old value
         disposeEntryValue(newEntryPtr);
       endif;

       // copy key and value to entry
       memcpy(newEntry.key : %addr(tmpKey) : keyLength);
       newEntry.value = objPtr;

       // set null to the last byte
       memcpy(newEntry.key + keyLength : %addr(hexNull) : 1);
      /end-free
     P                 E


     /**
      * \brief Get object entry
      *
      * Returns a pointer to the object.
      *
      * <br><br>
      *
      * If the entry is not of type object an escape message will be send.
      *
      * \author Mihael Schmidt
      * \date   11.05.2008
      *
      * \param Pointer to the JSON array
      * \param Index
      *
      * \return Pointer to the JSON object
      */
     P json_getObject...
     P                 B                   export
     D                 PI              *   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);
       tmpKey = %subst(key : 1 : length);

       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         if (entry.type = JSON_TYPE_OBJECT);
           retVal = entry.value;
         else;
           sendEscapeMessage(MSG_ENTRY_TYPE_MISMATCH);
         endif;
       endif;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Get entry type
      *
      * Returns the type of the entry.
      *
      * \param Pointer to the JSON object
      * \param Key
      *
      * \return Entry type
      */
     P json_getEntryType...
     P                 B                   export
     D                 PI            10I 0 opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D type            S             10I 0
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);

       tmpKey = %subst(key : 1 : length);
       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         type = entry.type;
       else;
         sendEscapeMessage(MSG_ENTRY_NOT_FOUND);
       endif;

       return type;
      /end-free
     P                 E


     /**
      * \brief Checks if entry is of special type null
      * 
      * \param Pointer to the JSON object
      * \param Key
      * 
      * \return *on = entry is of type null <br>
      *         *off = entry is of another type
      */
     P json_isNull     B                   export
     D                 PI              N   opdesc
     D   jsonPtr                       *   const
     D   key                      65535A   const options(*varsize)
      *
     D tmpKey          S          65535A
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
      /free
       isJSONImpl(jsonPtr);

       cee_getOpDescInfo(2 : descType : dataType : descInfo1 : descInfo2 :
                         length : *omit);

       tmpKey = %subst(key : 1 : length);
       entryPtr = getJSONEntry(jsonPtr : %addr(tmpKey) : length);
       if (entryPtr <> *null);
         return (entry.type = JSON_TYPE_NULL);
       else;
         sendEscapeMessage(MSG_ENTRY_NOT_FOUND);
         return *off; // dummy
       endif;
      /end-free
     P                 E
     

     /**
      * \brief Check for JSON object
      *
      * Checks if the passed pointer points to a JSON object.
      *
      * <br><br>
      *
      * If the pointer does not point to a JSON object and escape message will
      * be send to the next entry in the call stack.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to the JSON object
      */
     P isJSONImpl...
     P                 B                   export
     D                 PI
     D   jsonPtr                       *   const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
      /free
       monitor;
         if (header.id <> JSON_ID);
           sendEscapeMessage(MSG_NO_JSON_OBJECT);
         endif;

         on-error *all;
           sendEscapeMessage(MSG_NO_JSON_OBJECT);
       endmon;
      /end-free
     P                 E


     /**
      * \brief Send escape message
      *
      * A wrapper for the i5/OS API QMHSNDPM.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Message id
      */
     P sendEscapeMessage...
     P                 B                   export
     D                 PI
     D   id                          10I 0 const
      *
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D msgdata         S            512A
     D msgkey          S              4A
      /free
       if (id = MSG_NO_JSON_OBJECT);
         msgdata = 'The pointer does not point to a JSON object data structure';
       elseif (id = MSG_ENTRY_TYPE_MISMATCH);
         msgdata = 'The entry type does not match with the expected type.';
       elseif (id = MSG_ARRAY_ELEMENT_NOT_FOUND);
         msgdata = 'The specified array element was not found in the array.';
       elseif (id = MSG_TYPE_CAST_EXCEPTION);
         msgdata = 'The specified entry could not be cast to the requested ' +
                   'type.';
       else;
         return;
       endif;

       clear QUSEC;
       QUSBPRV = 0;
       sendProgramMessage('CPF9898' :
                          'QCPFMSG   *LIBL     ' :
                          %trimr(msgdata) :
                          %len(%trimr(msgdata)) :
                          '*ESCAPE   ' :
                          '*PGMBDY' :
                          0 :
                          msgkey :
                          QUSEC);
      /end-free
     P                 E


     /**
      * \brief Get entry by key
      *
      * Returns a pointer to the entry data structure for the passed key.
      *
      * \param Pointer to the JSON object
      * \param Pointer to the key
      * \param Key length
      *
      * \return Pointer to the entry or *null if no such entry exists in this
      *         JSON object.
      */
     P getJSONEntry    B                   export
     D                 PI              *
     D  jsonPtr                        *   const
     D  keyPtr                         *   const
     D  keyLength                    10U 0 const
      *
     D header          DS                  likeds(tmpl_header) based(jsonPtr)
     D entryPtr        S               *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
     D retVal          S               *   inz(*null)
      /free
       entryPtr = header.firstEntry;

       dow (entryPtr <> *null);

         if (keyLength = entry.keyLength - 1 and
             memcmp(keyPtr : entry.key : keyLength) = 0); // dont include the null
           retVal = entryPtr;
           leave;
         endif;

         entryPtr = entry.next;
       enddo;

       return retVal;
      /end-free
     P                 E


     /**
      * \brief Escape characters
      *
      * Double quotes and backslashes must be escape in string values to
      * comply to the JSON specs.
      *
      * <br><br>
      *
      * The last element of the token list will be checked for characters
      * which need to be escaped.
      *
      * \author Mihael Schmidt
      * \date   20.03.2008
      *
      * \param Pointer to token list
      */
     P escapeLastToken...
     P                 B                   export
     D                 PI
     D   tokenList                     *
      *
     D ptr             S               *
     D length          S             10I 0
     D backslash       S              1A   inz('\')
     D newPtr          S               *
     D newLength       S             10I 0
     D newCurLength    S             10I 0
     D i               S             10I 0
     D x               S             10I 0
     D iPtr            S               *
     D char            S              1A   based(iPtr)
      /free
       ptr = arraylist_getLast(tokenList);
       length = strlen(ptr);
       newLength = length;

       // number of character to be escaped
       for i = 0 to length - 1;
         iPtr = ptr + i;
         if (char = '\' or char = '"');
           newLength += 1;
         endif;
       endfor;

       // check if there are any characters which need to be escaped
       if (newLength <> length);
         newPtr = %alloc(newLength);

         for i = 0 to length - 1;
           iPtr = ptr + i;
           if (char = '\' or char = '"');
             memcpy(newPtr + newCurLength : ptr + x : i - x);
             newCurLength += i - x;
             // copy escape character
             memcpy(newPtr + newCurLength : %addr(backslash) : 1);
             newCurLength += 1;

             x = i; // save last position
           endif;
         endfor;

         // copy the rest of the string
         memcpy(newPtr + newCurLength : ptr + x : i - x + 1);
         newCurLength += i - x;

         // replace value in token list
         arraylist_removeLast(tokenList);
         arraylist_add(tokenList : newPtr : newCurLength);

         dealloc(n) newPtr;
       endif;
      /end-free
     P                 E


     /**
      * \brief Dealloc entry value
      *
      * The value of an entry is deallocated. If the entry is of
      * type object or array it must be disposed.
      *
      * \author Mihael Schmidt
      * \date   14.05.2008
      *
      * \param Pointer to an entry
      */
     P disposeEntryValue...
     P                 B                   export
     D                 PI
     D   entryPtr                      *
      *
     D entry           DS                  likeds(tmpl_entry) based(entryPtr)
      /free
       if (entry.type = JSON_TYPE_ARRAY);
         jsona_dispose(entry.value);
       elseif (entry.type = JSON_TYPE_OBJECT);
         json_dispose(entry.value);
       elseif (entry.type = JSON_TYPE_NULL);
         // do nothing, entry is already *null
       elseif (entry.value <> *null);
         dealloc(n) entry.value;
       endif;
      /end-free
     P                 E


     /**
      * \brief Send diagnostic message
      *
      * Sends a diagnostic message to the call stack
      * (one level above this one).
      *
      * \author Mihael Schmidt
      * \date   07.03.2009
      *
      * \param Message
      */
     P sendDiagnosticMessage...
     P                 B                   export
     D                 PI
     D   message                    200A   const
      *
     D sendProgramMessage...
     D                 PR                  extpgm('QMHSNDPM')
     D  szMsgID                       7A   const
     D  szMsgFile                    20A   const
     D  szMsgData                  6000A   const  options(*varsize)
     D  nMsgDataLen                  10I 0 const
     D  szMsgType                    10A   const
     D  szCallStkEntry...
     D                               10A   const
     D  nRelativeCallStkEntry...
     D                               10I 0 const
     D  szRtnMsgKey                   4A
     D  error                       265A   options(*varsize)
      *
     D msgdata         S            512A
     D msgkey          S              4A
      /free
       msgdata = message;

       clear QUSEC;
       QUSBPRV = 0;
       sendProgramMessage(*blank :
                          *blank :
                          %trimr(msgdata) :
                          %len(%trimr(msgdata)) :
                          '*DIAG' :
                          '*PGMBDY' :
                          0 :
                          msgkey :
                          QUSEC);
      /end-free
     P                 E


     P prettyPrint     B                   export
     D                 PI
     D   tokenList                     *   const
     D   printFormat                       const likeds(json_prettyPrintFormat)
     D   indentationLevel...
     D                               10I 0 const
      *
     D indentSpace     S           1000A   static inz
     D indentTab       S           1000A   static inz(*ALLx'05')
      /free
       if (printFormat.useTabs);
         arraylist_add(tokenList : %addr(indentTab) :
                  indentationLevel * printFormat.numberWhitespace);
       else;
         arraylist_add(tokenList : %addr(indentSpace) :
                  indentationLevel * printFormat.numberWhitespace);
       endif;
      /end-free
     P                 E


